﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Security.Cryptography;
using System.Windows.Forms;
using ff14bot.Enums;
using ff14bot.Navigation;
using ff14bot;
using ff14bot.Behavior;
using ff14bot.Helpers;
using ff14bot.Managers;
using ff14bot.Objects;
using TreeSharp;
using ff14bot.AClasses;


using Action = TreeSharp.Action;
using Eclipse.HealBot.Routine;
using Eclipse.HealBot;

namespace Eclipse.Healbot.Routine
{
    public abstract class EclipseRoutine : CombatRoutine
    {
        //public override virtual string Name { get { return "Eclipse.Healbot.Routine [" + GetType().Name + "]"; } }

        private Form _configForm;
        public override bool WantButton
        {
            get
            {
                return true;
            }
        }



        public static EclipseSettings WindowSettings = new EclipseSettings();

        public override void OnButtonPress()
        {
            if (_configForm == null || _configForm.IsDisposed || _configForm.Disposing)
                _configForm = new EclipseSettings();

            _configForm.ShowDialog();
        }

        #region CombatRoutine implementation

        private Composite _heal;



        public virtual void OnPulse()
        {
            
        }

        public sealed override void Pulse()
        {
            Extensions.DoubleCastPreventionDict.RemoveAll(t => DateTime.UtcNow > t);
            OnPulse();
        }

        public sealed override Composite HealBehavior
        {
            get { return _heal ?? (_heal = CreateHeal()); }
        }


        public static bool Resting = false;



        #region Hidden Overrides


        public override void Initialize()
        {
            RegisterHotkeys();

            //CombatLogHandler.Initialize();
        }

        public override void ShutDown()
        {
            UnregisterHotkeys();
            //CombatLogHandler.Shutdown();
        }

        #endregion

        #endregion

        #region Required Implementations

        protected virtual Composite CreateCombat()
        {
            return new HookExecutor("Eclipse_Combat_Root",
                "Root composite for Eclipse.Healbot.Routine combat. Rotations will be plugged into this hook.",
                new Action(r => RunStatus.Failure));
        }

        protected virtual Composite CreateRest()
        {
            return new HookExecutor("Eclipse_Rest_Root",
                "Root composite for Eclipse.Healbot.Routine Rest. Rotations will be plugged into this hook.",
                new Action(r => RunStatus.Failure));
        }

        protected virtual Composite CreatePreCombatBuffs()
        {
            return new HookExecutor("Eclipse_PreCombatBuffs_Root",
                "Root composite for Eclipse.Healbot.Routine buffs. Rotations will be plugged into this hook.",
                new Action(r => RunStatus.Failure));
        }

        protected virtual Composite CreateCombatBuffs()
        {
            return new HookExecutor("Eclipse_CombatBuffs_Root",
                "Root composite for Eclipse.Healbot.Routine buffs. Rotations will be plugged into this hook.",
                new Action(r => RunStatus.Failure));
        }

        protected virtual Composite CreateHeal()
        {
            EC.log("In CreateHeal method...");
            return new HookExecutor("Eclipse_Heals_Root",
                "Root composite for Eclipse.Healbot.Routine heals. Rotations will be plugged into this hook.",
                new Action(r => RunStatus.Failure));
        }

        protected virtual Composite CreatePull()
        {
            return new HookExecutor("Eclipse_Pull_Root",
                "Root composite for Eclipse.Healbot.Routine pull. Rotations will be plugged into this hook.",
                new Action(r => RunStatus.Failure));
        }

        protected bool InterruptsEnabled { get; set; }
        protected virtual void UnregisterHotkeys()
        {
            //HotkeysManager.Unregister("Eclipse.Healbot.Routine Toggle Interrupt");
        }
        protected virtual void RegisterHotkeys()
        {
            /*HotkeysManager.Register("Eclipse.Healbot.Routine Toggle Interrupt",
                Keys.NumPad1,
                ModifierKeys.Alt,
                o =>
                    {
                        InterruptsEnabled = !InterruptsEnabled;
                        Logging.Write("Interrupts enabled: " + InterruptsEnabled);
                    });

            // Default this to true please. Thanks!
            InterruptsEnabled = true;*/
        }

        #endregion

        #region Unit Wrappers
        protected BattleCharacter GetTargetMissingDebuff(GameObject near, string debuff, float distance = -1f)
        {
            return UnfriendlyUnits.FirstOrDefault(u => u.Location.Distance3D(near.Location) <= distance && !u.HasAura(debuff, true));
        }
        protected int EnemiesNearTarget(float range)
        {
            var target = Core.Player.CurrentTarget;
            if (target == null)
                return 0;
            var tarLoc = target.Location;
            return UnfriendlyUnits.Count(u => u.ObjectId != target.ObjectId && u.Location.Distance3D(tarLoc) <= range);
        }
        //protected IEnumerable<GameObject> UnfriendlyMeleeUnits { get { return UnfriendlyUnits.Where(u => Actionmanager.InSpellInRangeLOS()); } }

        protected IEnumerable<BattleCharacter> UnfriendlyUnits
        {
            get
            {
                return
                    GameObjectManager.GetObjectsOfType<BattleCharacter>().Where(u => !u.IsDead && u.CanAttack);
            }
        }

        #endregion



        #region Spell Casting
        protected Composite SummonChocobo()
        {
            return new PrioritySelector(


                new Decorator(r => !Chocobo.Summoned && Chocobo.CanSummon, new Action(r =>
                {
                    if (MovementManager.IsMoving)
                        Navigator.PlayerMover.MoveStop();

                    Chocobo.Summon();
                }))


                );
        }

        public Composite EnsureTarget
        {
            get
            {
                return new Decorator(r => Core.Player.SpellCastInfo != null && Core.Player.SpellCastInfo.TargetId != GameObjectManager.EmptyGameObject,
                    new PrioritySelector(
                        r => GameObjectManager.GetObjectByObjectId(Core.Player.SpellCastInfo.TargetId),
                        new Decorator(r => (r as Character) != null && (r as Character).IsDead,


                            new PrioritySelector(
                                new FailLogger(r => string.Format("Stop casting at {0} because it is dead.", (r as GameObject).Name)),
                                new Action(r => Actionmanager.StopCasting()))

                    )));
            }
        }

        protected Composite Cast(Selection<string> spell, Selection<bool> reqs = null, Selection<GameObject> onTarget = null, bool ignoreCanCast = false)
        {
            return new Decorator(
        ret =>
        {
            // Check reqs if its there.
            if (reqs != null && !reqs(ret))
                return false;

            // To ignore CanCast stuff (CanCast returns false while channeling, useful to be ignored for dots)
            if (!ignoreCanCast && !Actionmanager.CanCast(spell(ret), onTarget != null ? onTarget(ret) : Core.Player.CurrentTarget))
                return false;


            SpellData data;
            var result = Actionmanager.CurrentActions.TryGetValue(spell(ret), out data);
            if (result)
            {
                var castTime = data.AdjustedCastTime.TotalSeconds > 0;
                //Spells with a cast time should not be cast when moving unless we are autonomus inwhich case we want to stop
                if (castTime && MovementManager.IsMoving)
                {
                    if (BotManager.Current.IsAutonomous)
                    {
                        return true;
                    }
                    return false;
                }
            }


            return true;
        },
        new Action(ret =>
        {
            var castingSpell = Core.Player.SpellCastInfo;
            // If we're casting something other than what we should be, stop casting it. (Channeling /stopcast stuff)
            if (castingSpell != null && castingSpell.IsSpell && castingSpell.SpellData.Name != spell(ret))
                Actionmanager.StopCasting();

            if (MovementManager.IsMoving)
                Navigator.PlayerMover.MoveStop();


                Logging.Write("Casting " + spell(ret));
                Actionmanager.DoAction(spell(ret), (onTarget != null ? onTarget(ret) : Core.Player.CurrentTarget));
            
        }));
        }


        protected Composite Cast(string spell, Selection<bool> reqs = null, Selection<GameObject> onTarget = null, bool ignoreCanCast = false)
        {
            return
                new Decorator(
                    ret =>
                    {
                        // Check reqs if its there.
                        if (reqs != null && !reqs(ret))
                            return false;

                        // To ignore CanCast stuff (CanCast returns false while channeling, useful to be ignored for dots)
                        if (!ignoreCanCast && !Actionmanager.CanCast(spell, onTarget != null ? onTarget(ret) : Core.Player.CurrentTarget))
                            return false;


                        SpellData data;
                        var result = Actionmanager.CurrentActions.TryGetValue(spell, out data);
                        if (result)
                        {
                            var castTime = data.AdjustedCastTime.TotalSeconds > 0;
                            //Spells with a cast time should not be cast when moving unless we are autonomus inwhich case we want to stop
                            if (castTime && MovementManager.IsMoving)
                            {
                                if (BotManager.Current.IsAutonomous)
                                {
                                    return true;
                                }
                                return false;
                            }
                        }



                        return true;
                    },
                    new Action(ret =>
                    {
                        var castingSpell = Core.Player.SpellCastInfo;
                        // If we're casting something other than what we should be, stop casting it. (Channeling /stopcast stuff)
                        if (castingSpell != null && castingSpell.IsSpell && castingSpell.SpellData.Name != spell)
                            Actionmanager.StopCasting();


                        if (MovementManager.IsMoving)
                            Navigator.PlayerMover.MoveStop();


                            Logging.Write("Casting " + spell);
                            Actionmanager.DoAction(spell, (onTarget != null ? onTarget(ret) : Core.Player.CurrentTarget));
                        

                    }));
        }

        protected Composite Apply(Selection<string> spell, Selection<bool> reqs = null, Selection<GameObject> onTarget = null, int msLeft = 0, bool ignoreCanCast = false)
        {
            return new Decorator(ret =>
            {
                // Check reqs if its there.
                if (reqs != null && !reqs(ret))
                    return false;

                // Specific target, or just our general target.
                GameObject target = onTarget != null ? onTarget(ret) : Core.Player.CurrentTarget;


                if (target == null)
                    return false;

                // To ignore CanCast stuff (CanCast returns false while channeling, useful to be ignored for dots)
                if (!ignoreCanCast && !Actionmanager.CanCast(spell(ret), target))
                    return false;

                if (Extensions.DoubleCastPreventionDict.Contains(target, spell(ret)))
                    return false;



                

                if (!target.HasAura(spell(ret), true, msLeft))
                {


                    SpellData data;
                    var result = Actionmanager.CurrentActions.TryGetValue(spell(ret), out data);
                    if (result)
                    {
                        var castTime = data.AdjustedCastTime.TotalSeconds > 0;
                        //Spells with a cast time should not be cast when moving unless we are autonomus inwhich case we want to stop
                        if (castTime && MovementManager.IsMoving)
                        {
                            if (BotManager.Current.IsAutonomous)
                            {
                                return true;
                            }
                            return false;
                        }
                    }

 

                    return true;
                }





                return false;
            },
                new Action(ret =>
                {
                    var castingSpell = Core.Player.SpellCastInfo;
                    // If we're casting something other than what we should be, stop casting it. (Channeling /stopcast stuff)
                    if (castingSpell != null && castingSpell.IsSpell && castingSpell.SpellData.Name != spell(ret))
                        Actionmanager.StopCasting();

                    if (MovementManager.IsMoving)
                        Navigator.PlayerMover.MoveStop();





                    var unit = (onTarget != null ? onTarget(ret) : Core.Player.CurrentTarget);
                    Logging.Write("Applying " + spell(ret));
                    if (Actionmanager.DoAction(spell(ret), unit))
                    {
                        Extensions.UpdateDoubleCastDict(spell(ret), unit);
                    }



                }));
        }


        protected Composite Apply(string spell, Selection<bool> reqs = null, Selection<GameObject> onTarget = null, int msLeft = 0, bool ignoreCanCast = false)
        {
            return new Decorator(ret =>
                {
                    // Check reqs if its there.
                    if (reqs != null && !reqs(ret))
                        return false;

                    // Specific target, or just our general target.
                    GameObject target = onTarget != null ? onTarget(ret) : Core.Player.CurrentTarget;


                    if (target == null)
                        return false;

                    // To ignore CanCast stuff (CanCast returns false while channeling, useful to be ignored for dots)
                    if (!ignoreCanCast && !Actionmanager.CanCast(spell, target))
                        return false;

                    if (Extensions.DoubleCastPreventionDict.Contains(target, spell))
                        return false;

                    if (!target.HasAura(spell, true, msLeft))
                    {

                        SpellData data;
                        var result = Actionmanager.CurrentActions.TryGetValue(spell, out data);
                        if (result)
                        {
                            var castTime = data.AdjustedCastTime.TotalSeconds > 0;
                            //Spells with a cast time should not be cast when moving unless we are autonomus inwhich case we want to stop
                            if (castTime && MovementManager.IsMoving)
                            {
                                if (BotManager.Current.IsAutonomous)
                                {
                                    return true;
                                }
                                return false;
                            }
                        }




                        return true;
                    }

                    return false;
                },
                new Action(ret =>
                {
                    var castingSpell = Core.Player.SpellCastInfo;
                    // If we're casting something other than what we should be, stop casting it. (Channeling /stopcast stuff)
                    if (castingSpell != null && castingSpell.IsSpell && castingSpell.SpellData.Name != spell)
                        Actionmanager.StopCasting();


                    if (MovementManager.IsMoving)
                        Navigator.PlayerMover.MoveStop();


                        var unit = (onTarget != null ? onTarget(ret) : Core.Player.CurrentTarget);
                        Logging.Write("Applying " + spell);
                        if (Actionmanager.DoAction(spell, unit))
                        {
                            Extensions.UpdateDoubleCastDict(spell, unit);
                        }
                    


                }));
        }




        #endregion

        protected delegate T Selection<out T>(object context);
    }

    public static class Extensions
    {

        public static void RemoveAll<TKey, TValue>(this Dictionary<TKey, TValue> dic,
     Func<TValue, bool> predicate)
        {
            var keys = dic.Keys.Where(k => predicate(dic[k])).ToList();
            foreach (var key in keys)
            {
                dic.Remove(key);
            }
        }

        public static void UpdateDoubleCastDict(string spellName, GameObject unit)
        {
            if (unit == null)
                return;

            SpellData spellData;


            Actionmanager.CurrentActions.TryGetValue(spellName, out spellData);
            

            if (spellData == null)
                return;



            DateTime expir = DateTime.UtcNow + spellData.AdjustedCastTime + TimeSpan.FromSeconds(3);
            string key = DoubleCastKey(unit.ObjectId, spellName);
            if (DoubleCastPreventionDict.ContainsKey(key))
                DoubleCastPreventionDict[key] = expir;

            DoubleCastPreventionDict.Add(key, expir);
        }

        public static string DoubleCastKey(uint guid, string spellName)
        {
            return guid.ToString("X") + "-" + spellName;
        }

        public static string DoubleCastKey(GameObject unit, string spell)
        {
            return DoubleCastKey(unit.ObjectId, spell);
        }

        public static bool Contains(this Dictionary<string, DateTime> dict, GameObject unit, string spellName)
        {
            return dict.ContainsKey(DoubleCastKey(unit, spellName));
        }

        public static bool ContainsAny(this Dictionary<string, DateTime> dict, GameObject unit, params string[] spellNames)
        {
            return spellNames.Any(s => dict.ContainsKey(DoubleCastKey(unit, s)));
        }

        public static bool ContainsAll(this Dictionary<string, DateTime> dict, GameObject unit, params string[] spellNames)
        {
            return spellNames.All(s => dict.ContainsKey(DoubleCastKey(unit, s)));
        }

        public static readonly Dictionary<string, DateTime> DoubleCastPreventionDict = new Dictionary<string, DateTime>();



        public static bool HasAura(this GameObject unit, string spellname, bool isMyAura = false, int msLeft = 0)
        {
            var unitasc = (unit as Character);
            if (unit == null || unitasc == null)
            {
                return false;
            }
            IEnumerable<Aura> auras;
            auras = isMyAura ? unitasc.CharacterAuras.Where(r => r.CasterId == Core.Player.ObjectId && r.Name == spellname) : unitasc.CharacterAuras.Where(r => r.Name == spellname);
           

            return auras.Any(aura => aura.TimespanLeft.TotalMilliseconds > msLeft);
        }
    }
}